@@ -17,6 +17,16 @@ use cosmwasm_std::{Order, Pair};
 /// This is internal as it can change any time if the map implementation is swapped out.
 type BTreeMapPairRef<'a, T = Vec<u8>> = (&'a Vec<u8>, &'a T);
 
+pub fn transactional<F, T>(base: &mut dyn Storage, action: F) -> Result<T, String>
+where
+    F: FnOnce(&mut dyn Storage, &dyn Storage) -> Result<T, String>,
+{
+    let mut cache = StorageTransaction::new(base);
+    let res = action(&mut cache, base)?;
+    cache.prepare().commit(base);
+    Ok(res)
+}
+
 pub struct StorageTransaction<'a> {
     /// read-only access to backing storage
     storage: &'a dyn Storage,
@@ -39,10 +49,6 @@ impl<'a> StorageTransaction<'a> {
     pub fn prepare(self) -> RepLog {
         self.rep_log
     }
-
-    /// rollback will consume the checkpoint and drop all changes (not really needed, going out of scope does the same, but nice for clarity)
-    #[allow(dead_code)]
-    pub fn rollback(self) {}
 }
 
 impl<'a> Storage for StorageTransaction<'a> {
@@ -120,7 +126,7 @@ impl RepLog {
     }
 
     /// applies the stored list of `Op`s to the provided `Storage`
-    pub fn commit<S: Storage + ?Sized>(self, storage: &mut S) {
+    pub fn commit(self, storage: &mut dyn Storage) {
         for op in self.ops_log {
             op.apply(storage);
         }
@@ -142,7 +148,7 @@ enum Op {
 
 impl Op {
     /// applies this `Op` to the provided storage
-    pub fn apply<S: Storage + ?Sized>(&self, storage: &mut S) {
+    pub fn apply(&self, storage: &mut dyn Storage) {
         match self {
             Op::Set { key, value } => storage.set(&key, &value),
             Op::Delete { key } => storage.remove(&key),
@@ -567,19 +573,6 @@ mod test {
         assert_eq!(base.get(b"subtx"), Some(b"works".to_vec()));
     }
 
-    #[test]
-    fn rollback_has_no_effect() {
-        let mut base = MemoryStorage::new();
-        base.set(b"foo", b"bar");
-
-        let mut check = StorageTransaction::new(&base);
-        assert_eq!(check.get(b"foo"), Some(b"bar".to_vec()));
-        check.set(b"subtx", b"works");
-        check.rollback();
-
-        assert_eq!(base.get(b"subtx"), None);
-    }
-
     #[test]
     fn ignore_same_as_rollback() {
         let mut base = MemoryStorage::new();
